/* Sentinel - IRC Statistical and Operator Services
** util.c - Misc. utility functions
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: util.c,v 1.152 2003/04/17 13:08:28 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "match.h"
#include "lang.h"
#include <signal.h>

/* print a period of seconds into human readable format */
char *timediff(time_t u)
{
  static char buf[TIMEBUF];

  if (u > 86400)
  {
    time_t days = u / 86400;

    snprintf(buf, TIMEBUF, "%ld %s, %02ld:%02ld",
             (long)days, (days == 1) ? T_DAY : T_DAYS,
             (long)((u / 3600) % 24), (long)((u / 60) % 60));
  }
  else if (u > 3600)
  {
    time_t hours = u / 3600;
    time_t minutes = (u / 60) % 60;

    snprintf(buf, TIMEBUF, "%ld %s, %ld %s",
             (long)hours, (hours == 1) ? T_HOUR : T_HOURS, (long)minutes,
             (minutes == 1) ? T_MINUTE : T_MINUTES);
  }
  else
  {
    time_t minutes = u / 60;
    time_t seconds = u % 60;

    snprintf(buf, TIMEBUF,
             "%ld %s, %ld %s",
             (long)minutes, (minutes == 1) ? T_MINUTE : T_MINUTES,
             (long)seconds, (seconds == 1) ? T_SECOND : T_SECONDS);
  }

  return buf;
}

/* Our own ctime function */
/* Modified for l10n - 12 fv 2002 -Hwy */
char *sctime(time_t stuff)
{
  static char buf[32];
  struct tm *ltm = localtime(&stuff);
  static char *gcc_sucks = "%c";

  /* %c is the POSIX localized replacement for ctime() */

  /* gcc complains about some non-BSD implementations not using
  ** a 4 digit year in some locales.  The warning can be ignored.
  **
  ** The lameness in this function is the only way I know of that
  ** the warning can be ignored.
  **
  ** Some versions of gcc support -Wno-y2k, but apparently it's
  ** only gcc3 and higher.  Using a custom -W parameter will cause
  ** issues when run with another compiler, so for now, this lameness
  ** stays.
  */
  strftime(buf, 32, gcc_sucks, ltm);
  return buf;
}

/* custom ftime format */
char *sftime(time_t stuff)
{
  static char fmtime[LINE];
  struct tm *ltm = localtime(&stuff);

  if (stuff != 0)
  {
    strftime(fmtime, LINE, FTMFMT, ltm);
  }
  else
  {
    snprintf(fmtime, LINE, "No Time/Record");
  }

  return fmtime;
}

char *irc_time(time_t ts)
{
  static char fmtime[LINE];

  struct tm *ltm = localtime(&ts);

  /* Since we already depend on POSIX local support, %z in the time
  ** format string is acceptable...GMToffset was a nice function
  ** though :)
  */
  strftime(fmtime, LINE, IRCTFMT, ltm);

  return fmtime;
}

/* tokenize splits apart a message (in our case with the origin and command
** picked off already...
*/
int tokenize(char *message, char **parv)
{
  char *pos = NULL;
  char *next;
  int count = 0;

  if (!message)
  {
    /* Something is seriously wrong...bail */
    return -1;
  }

  /* First we find out if there's a : in the message, save that string
  ** somewhere so we can set it to the last param in parv
  ** Also make sure there's a space before it...if not, then we're
  ** screwed
  */
  pos = message;
  while (TRUE)
  {
    if ((pos = strchr(pos, ':')))
    {
      pos--;
      if (*pos != ' ')
      {
        /* There SHOULDN'T be a problem with this... */
        pos += 2;
        continue;
      }
      *pos = '\0';
      /* This is BAD - get rid of the space before the : */
      pos++;
      *pos = '\0';
      /* VERY BAD...over now though */
      pos++;
      break;
    }
    else
    {
      break;
    }
  }

  /* Now we take the beginning of message and find all the spaces...
  ** set them to \0 and use 'next' to go through the string
  */

  next = message;
  parv[0] = message;
  count = 1;

  while (*next)
  {
    if (count == MAXPARA - 1)
    {
      /* We've reached one less than our max limit
      ** to handle the parameter we already ripped off
      */
      debug("Reached the MAXPARA limit!");
      return count;
    }
    if (*next == ' ')
    {
      *next = '\0';
      next++;
      /* Eat any additional spaces */
      while (*next == ' ')
        next++;
      /* If it's the end of the string, it's simply
      ** an extra space before the :parameter.  Here we
      ** break.
      */
      if (*next == '\0')
        break;
      parv[count] = next;
      count++;
    }
    else
    {
      next++;
    }
  }

  if (pos)
  {
    parv[count] = pos;
    count++;
  }

  return count;
}

/* A generic tokenizer, tokenizes based on a specified character */
int generictoken(char delim, int size, char *message, char **parv)
{
  char *next;
  int count;

  if (!message)
  {
    /* Something is seriously wrong...bail */
    return -1;
  }

  /* Now we take the beginning of message and find all the spaces...
  ** set them to \0 and use 'next' to go through the string
  */
  next = message;
  parv[0] = next;
  count = 1;

  while (*next)
  {
    /* This is fine here, since we don't have a :delimited
    ** parameter like tokenize
    */
    if (count == size)
    {
      /* We've reached our limit */
      debug("Reached the MAXPARA limit!");
      return count;
    }
    if (*next == delim)
    {
      *next = '\0';
      next++;
      /* Eat any additional delimiters */
      while (*next == delim)
        next++;
      /* If it's the end of the string, it's simply
      ** an extra space at the end.  Here we break.
      */
      if (*next == '\0')
        break;
      parv[count] = next;
      count++;
    }
    else
    {
      next++;
    }
  }

  return count;
}

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS) || defined(IRCD_BAHAMUT) 
/* A specific tokenizer for SJOIN, since Hybrid 7 likes adding extra spaces */
int sjtoken(char *message, char **parv)
{
  char *next;
  int count;

  if (!message)
  {
    /* Something is seriously wrong...bail */
    return -1;
  }

  /* Now we take the beginning of message and find all the spaces...
  ** set them to \0 and use 'next' to go through the string
  */
  next = message;

  /* Eat any initial delimiters */
  while (*next == ' ')
    next++;

  parv[0] = next;
  count = 1;

  while (*next)
  {
    /* This is fine here, since we don't have a :delimited
    ** parameter like tokenize
    */
    if (count == SJPARA)
    {
      /* We've reached our limit */
      debug("Reached the SJPARA limit!");
      return count;
    }
    if (*next == ' ')
    {
      *next = '\0';
      next++;
      /* Eat any additional delimiters */
      while (*next == ' ')
        next++;
      /* If it's the end of the string, it's simply
      ** an extra space at the end.  Here we break.
      */
      if (*next == '\0')
        break;
      /* If it happens to be a stray \r, break too */
      if (*next == '\r')
        break;
      parv[count] = next;
      count++;
    }
    else
    {
      next++;
    }
  }

  return count;
}
#endif

/* our own malloc with logging, sanity checking, and zeroing out memory */
void *smalloc(size_t size)
{
  void *buf;

  if (!size)
  {
    log(L_CRIT, "smalloc(): illegal attempt to allocate 0 bytes!");
    size = 1;
  }
  buf = calloc(1, size);
  if (!buf)
  {
    log(L_ALERT, "smalloc(): out of memory.");
    close_logfiles();
    exit(0);
  }
  return buf;
}

/* Generate a hash value into the tables */
unsigned long HASH(const unsigned char *name, int size_of_table)
{
  unsigned long h = 0, g;

  while (*name)
  {
    h = (h << 4) + ToLower(*name++);
    if ((g = (h & 0xF0000000)))
      h ^= g >> 24;
    h &= ~g;
  }
  return h % size_of_table;
}

/* Get the version flags string */
char *VersionFlags()
{
  static char ver[LINE];
  char *verptr = ver;

#ifdef TRACK_ADMIN

  *verptr++ = 'A';
#endif
#ifdef SUPPORT_MSG_UH

  *verptr++ = 'U';
#endif
#ifdef SUPPORT_MSG_OPERS

  *verptr++ = 'O';
#endif
#ifdef TRACK_VERSION

  *verptr++ = 'V';
#endif
#ifdef JUPESERV

  *verptr++ = 'J';
#endif
#ifdef ENABLE_QUOTE

  *verptr++ = 'Q';
#endif
#ifdef ENABLE_OP

  *verptr++ = 'o';
#endif
#ifdef ENABLE_TAKEOVER

  *verptr++ = 'K';
#endif
#ifdef TRACK_GECOS

  *verptr++ = 'G';
#endif
#ifdef HIDE_IPS

  *verptr++ = 'H';
#endif
#ifdef SUPPORT_HOSTSERV

  *verptr++ = 'h';
#endif
#ifdef OPER_MODES

  *verptr++ = 'M';
#endif
#ifdef DIE_ON_MASS_COLLIDE

  *verptr++ = 'C';
#endif
#ifdef USE_CURRENT_TS

  *verptr++ = 'e';
#endif
#ifdef DISALLOW_NICK_PRIVMSG

  *verptr++ = 'u';
#endif
#ifdef SPLITSERV

  *verptr++ = 'P';
#endif
#ifdef PUBSTATS

  *verptr++ = 'p';
#endif
#ifdef STATSERV

  *verptr++ = 'T';
#endif
#ifdef MONSERV

  *verptr++ = 'm';
#endif
#if defined(MONSERV) && defined(MON_IDENT_CLONES)

  *verptr++ = 'c';
#endif
#if defined(MONSERV) && defined(MON_DRONES)

  *verptr++ = 't';
#endif
#ifdef ALERTMODE

  *verptr++ = 'a';
#endif
#ifdef SPYMODE

  *verptr++ = 's';
#endif

  if (IsDebug())
    *verptr++ = 'd';
  if (IsNoFork())
    *verptr++ = 'n';

  *verptr = '\0';

#ifdef USE_ASCII_CASEMAP

  strlcat(ver, "(ASCII) ", LINE);
#else

  strlcat(ver, "(RFC1459) ", LINE);
#endif

#ifdef IRCD_HYBRID

  strlcat(ver, "HYBRID", LINE);
#endif
#ifdef IRCD_HYB6

  strlcat(ver, "HYB6", LINE);
#endif
#ifdef IRCD_HYB7

  strlcat(ver, "HYB7", LINE);
#endif
#ifdef IRCD_2_10

  strlcat(ver, "2_10", LINE);
#endif
#ifdef IRCD_BAHAMUT

  strlcat(ver, "BAHAMUT", LINE);
#endif
#ifdef IRCD_HYBSVS
#ifdef IRCD_HYBBG

  strlcat(ver, "HYBBG", LINE);
#else

  strlcat(ver, "HYBSVS", LINE);
#endif
#endif

  return ver;
}

#ifdef OPER_MODES
char opermode(struct User *u)
{
  if (u == NULL)
    return '-';
  if (u->operflags & OPER_MYOPER)
    return '*';
  if (u->operflags & OPER_SVSADMIN)
    return 'a';
  if (u->operflags & OPER_ADMIN)
    return 'A';
  if (u->operflags & OPER_GLOBAL)
    return 'O';
  if (u->operflags & OPER_LOCAL)
    return 'o';
  /* Shouldn't get here */
  return 'x';
}
#endif /* OPER_MODES */

void SigHup(int sig)
{
  need_rehash = 1;
  signal(SIGHUP, SigHup);
}

void SigUsr1(int sig)
{
  /* This should be sufficient to tell it to reconnect */
  close(servsock);
  delete_connection(servsock);
  servsock = -1;
  ServTermination();
  signal(SIGUSR1, SigUsr1);
}

void SigTerm(int sig)
{
  wallops(settings.name, "Received SIGTERM: Exiting");
  alert("Received SIGTERM: Exiting");
  sendto_dcc_flag(CF_NOTICES, "Received SIGTERM: Exiting");
#ifdef STATSERV

  sts(":%s QUIT :Received SIGTERM: Exiting", StatServ.nick);
#endif

  sts(":%s QUIT :Received SIGTERM: Exiting", AdminServ.nick);
#ifdef SPLITSERV

  sts(":%s QUIT :Received SIGTERM: Exiting", SplitServ.nick);
#endif
#ifdef JUPESERV

  sts(":%s QUIT :Received SIGTERM: Exiting", JupeServ.nick);
#endif
#ifdef MONSERV

  sts(":%s QUIT :Received SIGTERM: Exiting", MonServ.nick);
#endif

  sleep(1);
  close(servsock);
  close_dccs();
  SaveStats();
  log(L_ALERT, "Received SIGTERM: Exiting");
  close_logfiles();
  /* No need to reinstall the signal handler */
  exit(0);
}

void SigInt(int sig)
{
  wallops(settings.name, "Received SIGINT: Exiting");
  alert("Received SIGINT: Exiting");
  sendto_dcc_flag(CF_NOTICES, "Received SIGINT: Exiting");
#ifdef STATSERV

  sts(":%s QUIT :Received SIGINT: Exiting", StatServ.nick);
#endif

  sts(":%s QUIT :Received SIGINT: Exiting", AdminServ.nick);
#ifdef SPLITSERV

  sts(":%s QUIT :Received SIGINT: Exiting", SplitServ.nick);
#endif
#ifdef JUPESERV

  sts(":%s QUIT :Received SIGINT: Exiting", JupeServ.nick);
#endif
#ifdef MONSERV

  sts(":%s QUIT :Received SIGINT: Exiting", MonServ.nick);
#endif

  sleep(1);
  close(servsock);
  close_dccs();
  SaveStats();
  log(L_ALERT, "Received SIGINT: Exiting");
  close_logfiles();
  /* No need to reinstall the signal handler */
  exit(0);
}

int is_channel(char *chan)
{
  if (chan != NULL)
  {
    /* Sadly, a few servers propogate &channels... CSr comes to
    ** mind here.
    */
    if ( (chan[0] == '#') || (chan[0] == '+') || (chan[0] == '&')
#ifdef IRCD_2_10
         || (chan[0] == '!')
#endif
       )
      return TRUE;
    return FALSE;
  }

  /* Shouldn't get here */
  return FALSE;
}

void alert(char *fmt, ...)
{
#ifdef ALERTMODE
  va_list ap;
  char buf[BUFSIZE];

  /* This is simply a crutch... */
  if (me.s == NULL)
    return ;
  if (ad_delayed)
    return ;
  va_start(ap, fmt);
  vsnprintf(buf, sizeof(buf), fmt, ap);
  /* snprintf will always NULL terminate */
  privmsg(ALERTCHAN, AdminServ.nick, MSG_PRIVMSG, "%s", buf);
  va_end(ap);
#endif
}

#ifdef DRONES_CHANNEL
void dronealert(char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE];

  if (me.s == NULL)
    return ;
  if (ad_delayed)
    return ;
  va_start(ap, fmt);
  vsnprintf(buf, sizeof(buf), fmt, ap);
  /* snprintf will always NULL terminate */
  privmsg(DRONES_CHANNEL, AdminServ.nick, MSG_PRIVMSG, "%s", buf);
  va_end(ap);
}
#endif

void spy(char *fmt, ...)
{
#ifdef SPYMODE
  va_list ap;
  char buf[BUFSIZE];

  if (me.s == NULL)
    return ;
  if (ad_delayed)
    return ;
  va_start(ap, fmt);
  vsnprintf(buf, sizeof(buf), fmt, ap);
  privmsg(SPYCHAN, AdminServ.nick, MSG_PRIVMSG, "%s", buf);
  va_end(ap);
#endif
}

void debug(char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE];

  if (IsDebug())
  {
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    log(L_DEBUG, "DEBUG: %s", buf);
    va_end(ap);
  }
}

void InitLists(void)
{
  hiddenhead.head = NULL;
  hiddenhead.tail = NULL;
#ifdef IRCD_2_10

  servicehead.head = NULL;
  servicehead.tail = NULL;
#endif

  global_oper_list.head = NULL;
  global_oper_list.tail = NULL;
#if defined(MONSERV) && defined(MON_DRONES)

  MonServ.ignore_server_list.head = NULL;
  MonServ.ignore_server_list.tail = NULL;
  MonServ.gecos_drone_list.head = NULL;
  MonServ.gecos_drone_list.tail = NULL;
  MonServ.nick_drone_list.head = NULL;
  MonServ.nick_drone_list.tail = NULL;
  MonServ.channel_drone_list.head = NULL;
  MonServ.channel_drone_list.tail = NULL;
  MonServ.username_drone_list.head = NULL;
  MonServ.username_drone_list.tail = NULL;
  MonServ.drone_users.head = NULL;
  MonServ.drone_users.tail = NULL;
#endif
#ifdef ENABLE_TAKEOVER

  takeover_list.head = NULL;
  takeover_list.tail = NULL;
#endif
#ifdef JUPESERV

  server_jupes.head = NULL;
  server_jupes.tail = NULL;
  nick_jupes.head = NULL;
  nick_jupes.tail = NULL;
#endif
  connection_list.head = NULL;
  connection_list.tail = NULL;
  dying_list.head = NULL;
  dying_list.tail = NULL;
}

void ExpandParv(char *dest, int max, int initial, int parc, char *parv[])
{
  int i;

  if (parv[initial])
    strlcpy(dest, parv[initial], max);
  for (i = initial + 1; i < parc; i++)
  {
    strlcat(dest, " ", max);
    strlcat(dest, parv[i], max);
  }
}

/* Is it the midnight hour */
int is_midnight()
{
  time_t current = time(NULL);
  struct tm *ltm = localtime(&current);

  if (ltm->tm_hour == 0)
    return 1;

  return 0;
}

/* Perform a 3/4 of a second delay using select() */
void select_delay(void)
{
  struct timeval tv;

  tv.tv_usec = 750000; /* 3/4 a second delay */
  tv.tv_sec = 0;       /* ignore this */
  select(1, NULL, NULL, NULL, &tv);
}

/* Return 1 if the string s is all lower case, return 0 if not */
int strislower(char *s)
{
  char *t;

  if (s == NULL)
    return 1;
  for (t = s; *t != '\0'; t++)
  {
    if (islower((int) *t) == 0)
    {
      return 0;
    }
  }
  return 1;
}

/* Remove CR and LF from the line input.  ASSUME that the line
** does NOT contain CR or LF in the middle
*/
void strip(char *line)
{
  char *c;
  int len;

  if ((c = strchr(line, '\r')))
  {
    /* If we receive a CR, then we can just terminate the
    ** call there.
    **
    ** Not any longer :(  - We have to pare off the spaces
    ** at the end of lines
    */
    *c = '\0';
  }

  if ((c = strchr(line, '\n')))
  {
    *c = '\0';
  }

  /* Strip out trailing spaces, which can cause odd issues with
  ** SplitServ's channel and other things
  */
  len = strlen(line);
  while (len > 0)
  {
    if (line[len - 1] == ' ')
    {
      line[len - 1] = '\0';
      len--;
    }
    else
    {
      break;
    }
  }

  return ;
}

